import logging
import pandas as pd
import matplotlib.pyplot as plt
import statsmodels.formula.api as sm

from ..calculator import Calculator
from ..utils import get_extension, set_chart_style

from .cycletime import CycleTimeCalculator

logger = logging.getLogger(__name__)

class ThroughputCalculator(Calculator):
    """Build a data frame with columns `completed_timestamp` of the
    given frequency, and `count`, where count is the number of items
    completed at that timestamp (e.g. daily).
    """

    def run(self):
        cycle_data = self.get_result(CycleTimeCalculator)
        frequency = self.settings['throughput_frequency']
        
        logger.debug("Calculating throughput at frequency %s", frequency)

        if len(cycle_data.index) == 0:
            return pd.DataFrame([], columns=['count'], index=[])

        return cycle_data[['completed_timestamp', 'key']] \
            .rename(columns={'key': 'count'}) \
            .groupby('completed_timestamp').count() \
            .resample(frequency).sum() \
            .fillna(0)
    
    def write(self):
        data = self.get_result()
        
        if self.settings['throughput_data']:
            self.write_file(data, self.settings['throughput_data'])
        else:
            logger.debug("No output file specified for throughput data")
        
        if self.settings['throughput_chart']:
            self.write_chart(data, self.settings['throughput_chart'])
        else:
            logger.debug("No output file specified for throughput chart")

    def write_file(self, data, output_file):
        output_extension = get_extension(output_file)

        logger.info("Writing throughput data to %s", output_file)
        if output_extension == '.json':
            data.to_json(output_file, date_format='iso')
        elif output_extension == '.xlsx':
            data.to_excel(output_file, 'Throughput', header=True)
        else:
            data.to_csv(output_file, header=True)
    
    def write_chart(self, data, output_file):
        chart_data = data.copy()

        if len(chart_data.index) == 0:
            logger.warning("Cannot draw throughput chart with no completed items")
            return
        
        fig, ax = plt.subplots()

        if self.settings['throughput_chart_title']:
            ax.set_title(self.settings['throughput_chart_title'])

        window = self.settings['throughput_window']
        if window:
            chart_data = chart_data[-window:]

        # Calculate zero-indexed days to allow linear regression calculation
        day_zero = chart_data.index[0]
        chart_data['day'] = (chart_data.index - day_zero).days

        # Fit a linear regression (http://stackoverflow.com/questions/29960917/timeseries-fitted-values-from-trend-python)
        fit = sm.ols(formula="count ~ day", data=chart_data).fit()
        chart_data['fitted'] = fit.predict(chart_data)

        # Plot

        ax.set_xlabel("Period starting")
        ax.set_ylabel("Number of items")

        ax.plot(chart_data.index, chart_data['count'], marker='o')
        plt.xticks(chart_data.index, [d.date().strftime('%d/%m/%Y') for d in chart_data.index], rotation=70, size='small')

        _, top = ax.get_ylim()
        ax.set_ylim(0, top + 1)

        for x, y in zip(chart_data.index, chart_data['count']):
            if y == 0:
                continue
            ax.annotate(
                "%.0f" % y,
                xy=(x.toordinal(), y + 0.2),
                ha='center',
                va='bottom',
                fontsize="x-small",
            )

        ax.plot(chart_data.index, chart_data['fitted'], '--', linewidth=2)

        set_chart_style()

        # Write file
        logger.info("Writing throughput chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)



import pytest
from pandas import DataFrame, Timestamp

from .cycletime import CycleTimeCalculator
from .cfd import CFDCalculator

from ..utils import extend_dict

@pytest.fixture
def query_manager(minimal_query_manager):
    return minimal_query_manager

@pytest.fixture
def settings(minimal_settings):
    return extend_dict(minimal_settings, {})

@pytest.fixture
def columns(minimal_cycle_time_columns):
    return minimal_cycle_time_columns

def test_empty(query_manager, settings, columns):
    results = {
        CycleTimeCalculator: DataFrame([], columns=columns)
    }

    calculator = CFDCalculator(query_manager, settings, results)

    data = calculator.run()
    assert len(data.index) == 0

def test_columns(query_manager, settings, minimal_cycle_time_results):
    calculator = CFDCalculator(query_manager, settings, minimal_cycle_time_results)

    data = calculator.run()

    assert list(data.columns) == [
        'Backlog',
        'Committed',
        'Build',
        'Test',
        'Done'
    ]

def test_calculate_cfd(query_manager, settings, minimal_cycle_time_results):
    calculator = CFDCalculator(query_manager, settings, minimal_cycle_time_results)

    data = calculator.run()

    assert list(data.index) == [
        Timestamp('2018-01-01 00:00:00', freq='D'),
        Timestamp('2018-01-02 00:00:00', freq='D'),
        Timestamp('2018-01-03 00:00:00', freq='D'),
        Timestamp('2018-01-04 00:00:00', freq='D'),
        Timestamp('2018-01-05 00:00:00', freq='D'),
        Timestamp('2018-01-06 00:00:00', freq='D')
    ]

    assert data.to_dict('records') == [
        {'Backlog': 1.0, 'Committed': 0.0, 'Build': 0.0, 'Test': 0.0, 'Done': 0.0},
        {'Backlog': 2.0, 'Committed': 0.0, 'Build': 0.0, 'Test': 0.0, 'Done': 0.0},
        {'Backlog': 3.0, 'Committed': 2.0, 'Build': 0.0, 'Test': 0.0, 'Done': 0.0},
        {'Backlog': 4.0, 'Committed': 3.0, 'Build': 1.0, 'Test': 0.0, 'Done': 0.0},
        {'Backlog': 4.0, 'Committed': 3.0, 'Build': 1.0, 'Test': 1.0, 'Done': 0.0},
        {'Backlog': 4.0, 'Committed': 3.0, 'Build': 1.0, 'Test': 1.0, 'Done': 1.0},
    ]



import logging

from ..calculator import Calculator
from ..utils import get_extension

from .cycletime import CycleTimeCalculator

logger = logging.getLogger(__name__)

class PercentilesCalculator(Calculator):
    """Build percentiles for `cycle_time` in cycle data as a DataFrame
    """

    def run(self):
        cycle_data = self.get_result(CycleTimeCalculator)

        quantiles = self.settings['quantiles']
        logger.debug("Calculating percentiles at %s", ', '.join(['%.2f' % (q * 100.0) for q in quantiles]))

        return cycle_data['cycle_time'].dropna().quantile(quantiles)

    def write(self):
        output_file = self.settings['percentiles_data']
        if not output_file:
            logger.debug("No output file specified for percentiles data")
            return

        output_extension = get_extension(output_file)

        file_data = self.get_result()

        logger.info("Writing percentiles data to %s", output_file)
        if output_extension == '.json':
            file_data.to_json(output_file, date_format='iso')
        elif output_extension == '.xlsx':
            file_data.to_frame(name='percentiles').to_excel(output_file, 'Percentiles', header=True)
        else:
            file_data.to_csv(output_file, header=True)



import logging
import datetime

import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.transforms

from ..calculator import Calculator
from ..utils import set_chart_style, to_days_since_epoch

from .cycletime import CycleTimeCalculator
from .burnup import BurnupCalculator

logger = logging.getLogger(__name__)

class BurnupForecastCalculator(Calculator):
    """Draw a burn-up chart with a forecast run to completion
    """

    def run(self):
        burnup_data = self.get_result(BurnupCalculator)
        cycle_data = self.get_result(CycleTimeCalculator)

        if len(cycle_data.index) == 0:
            return None

        # This calculation is expensive. Only run it if we intend to write a file.
        if not self.settings['burnup_forecast_chart']:
            logger.debug("Not calculating burnup forecast chart data as no output file specified")
            return None

        backlog_column = self.settings['backlog_column']
        done_column = self.settings['done_column']

        if backlog_column not in burnup_data.columns:
            logger.error("Backlog column %s does not exist", backlog_column)
            return None
        if done_column not in burnup_data.columns:
            logger.error("Backlog column %s does not exist", done_column)
            return None

        if cycle_data[done_column].max() is pd.NaT:
            logger.warning("Unable to draw burnup forecast chart with zero completed items.")
            return None
        
        throughput_window_end = self.settings['burnup_forecast_chart_throughput_window_end'] or cycle_data[done_column].max().date()
        throughput_window = self.settings['burnup_forecast_chart_throughput_window']
        throughput_window_start = throughput_window_end - datetime.timedelta(days=throughput_window)
        logger.info("Sampling throughput between %s and %s", throughput_window_start.isoformat(), throughput_window_end.isoformat())

        target = self.settings['burnup_forecast_chart_target'] or burnup_data[backlog_column].max()
        logger.info("Running forecast to completion of %d items", target)
        
        trials = self.settings['burnup_forecast_chart_trials']
        logger.debug("Running %d trials to calculate probable forecast outcomes", trials)
        
        # calculate daily throughput
        throughput_data = cycle_data[
            (cycle_data[done_column] >= throughput_window_start) &
            (cycle_data[done_column] <= throughput_window_end)
        ][[done_column, 'key']] \
            .rename(columns={'key': 'count', done_column: 'completed_timestamp'}) \
            .groupby('completed_timestamp').count() \
            .resample("1D").sum() \
            .reindex(index=pd.DatetimeIndex(start=throughput_window_start, end=throughput_window_end, freq='D')) \
            .fillna(0)
        
        return burnup_monte_carlo(
            start_value=burnup_data[done_column].max(),
            target_value=target,
            start_date=burnup_data.index.max(),
            throughput_data=throughput_data,
            trials=trials
        )
    
    def write(self):
        output_file = self.settings['burnup_forecast_chart']
        if not output_file:
            logger.debug("No output file specified for burnup forecast chart")
            return
        
        burnup_data = self.get_result(BurnupCalculator)
        if burnup_data is None or len(burnup_data.index) == 0:
            logger.warning("Cannot draw burnup forecast chart with zero items")
            return
        
        window = self.settings['burnup_forecast_window']
        if window:
            start = burnup_data.index.max() - pd.Timedelta(window, 'D')
            burnup_data = burnup_data[start:]

            if len(burnup_data.index) == 0:
                logger.warning("Cannot draw burnup forecast chart with zero items")
                return
        
        mc_trials = self.get_result()
        if mc_trials is None:
            logger.warning("Cannot draw burnup forecast chart with zero completed trials")
            return

        deadline = self.settings['burnup_forecast_chart_deadline']
        if deadline:
            logger.debug("Forecasting with deadline", deadline.isoformat())

        deadline_confidence = self.settings['burnup_forecast_chart_deadline_confidence']
        if deadline_confidence:
            logger.debug("Forecasting deadline at %.2f%% confidence", deadline_confidence * 100.0)

        quantiles = self.settings['quantiles']
        logger.debug("Showing forecast at quantiles %s", ', '.join(['%.2f' % (q * 100.0) for q in quantiles]))
        
        backlog_column = self.settings['backlog_column']
        target = self.settings['burnup_forecast_chart_target'] or burnup_data[backlog_column].max()

        fig, ax = plt.subplots()
        
        if self.settings['burnup_forecast_chart_title']:
            ax.set_title(self.settings['burnup_forecast_chart_title'])

        fig.autofmt_xdate()

        transform_vertical = matplotlib.transforms.blended_transform_factory(ax.transData, ax.transAxes)
        transform_horizontal = matplotlib.transforms.blended_transform_factory(ax.transAxes, ax.transData)

        # plot backlog and burnup to date
        burnup_data.plot.line(ax=ax, legend=False)
        
        deadline_confidence_date = None

        # plot each monte carlo simulation line
        if mc_trials is not None:

            for col in mc_trials:
                mc_trials[col][mc_trials[col] > target] = target

            mc_trials.plot.line(ax=ax, legend=False, color='#ff9696', linestyle='solid', linewidth=0.1)

            # draw quantiles at finish line
            finish_dates = mc_trials.apply(pd.Series.last_valid_index)
            finish_date_quantiles = finish_dates.quantile(quantiles).dt.normalize()
            
            if deadline_confidence is not None:
                deadline_confidence_quantiles = finish_dates.quantile([deadline_confidence]).dt.normalize()
                if len(deadline_confidence_quantiles) > 0:
                    deadline_confidence_date = pd.Timestamp(deadline_confidence_quantiles.values[0]).to_pydatetime().date()

            bottom, top = ax.get_ylim()
            for percentile, value in finish_date_quantiles.iteritems():
                ax.vlines(value, bottom, target, linestyles='--', linewidths=0.5)
                ax.annotate("%.0f%% (%s)" % ((percentile * 100), value.strftime("%d/%m/%Y"),),
                    xy=(to_days_since_epoch(value.to_pydatetime().date()), 0.35),
                    xycoords=transform_vertical,
                    rotation="vertical",
                    ha="left",
                    va="top",
                    fontsize="x-small",
                    backgroundcolor="#ffffff"
                )
        
        # draw deadline (pun not intended...)
        if deadline is not None:
            bottom, top = ax.get_ylim()
            left, right = ax.get_xlim()
            
            deadline_dse = to_days_since_epoch(deadline)

            ax.vlines(deadline, bottom, target, color='r', linestyles='-', linewidths=0.5)
            ax.annotate("Due: %s" % (deadline.strftime("%d/%m/%Y"),),
                xy=(deadline, target),
                xytext=(0.95, 0.95),
                textcoords='axes fraction',
                arrowprops={
                    'arrowstyle': '->',
                    'color': 'r',
                    'linewidth': 1.1,
                    'connectionstyle': 'arc3,rad=.1',
                },
                fontsize="x-small",
                ha="right",
                color='red',
                backgroundcolor="#ffffff"
            )
            
            # Make sure we can see deadline line
            if right < deadline_dse:
                ax.set_xlim(left, deadline_dse + 1)
            
            # Draw deadline warning
            if deadline_confidence_date is not None:
                deadline_delta = (deadline - deadline_confidence_date).days
                
                ax.text(0.02, 0.5,
                    "Deadline: %s\nForecast (%.0f%%): %s\nSlack: %d days" % (
                        deadline.strftime("%d/%m/%Y"),
                        (deadline_confidence * 100),
                        deadline_confidence_date.strftime("%d/%m/%Y"),
                        deadline_delta
                    ),
                    transform=ax.transAxes,
                    fontsize=14,
                    verticalalignment='center',
                    bbox=dict(boxstyle='round', facecolor='r' if deadline_delta < 0 else 'g', alpha=0.5),
                )

        # Place target line
        left, right = ax.get_xlim()
        ax.hlines(target, left, right, linestyles='--', linewidths=1)
        ax.annotate("Target: %d" % (target,),
            xy=(0.02, target),
            xycoords=transform_horizontal,
            fontsize="x-small",
            ha="left",
            va="center",
            backgroundcolor="#ffffff"
        )
        
        # Give some headroom above the target line so we can see it
        bottom, top = ax.get_ylim()
        ax.set_ylim(bottom, int(top * 1.05))

        # Place legend underneath graph
        box = ax.get_position()
        handles, labels = ax.get_legend_handles_labels()
        ax.set_position([box.x0, box.y0 + box.height * 0.1, box.width, box.height * 0.9])

        ax.legend(handles[:2], labels[:2], loc='upper center', bbox_to_anchor=(0.5, -0.2), ncol=2)
        
        set_chart_style()

        # Write file
        logger.info("Writing burnup forecast chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)

def burnup_monte_carlo(start_value, target_value, start_date, throughput_data, trials=100):

    frequency = throughput_data.index.freq

    # degenerate case - no steps, abort
    if throughput_data['count'].sum() <= 0:
        logger.warning("No throughput samples available, aborting forecast simulations")
        return None

    # guess how far away we are; drawing samples one at a time is slow
    sample_buffer_size = int(2 * (target_value - start_value) / throughput_data['count'].mean())

    sample_buffer = dict(idx=0, buffer=None)

    def get_sample():
        if sample_buffer['buffer'] is None or sample_buffer['idx'] >= len(sample_buffer['buffer'].index):
            sample_buffer['buffer'] = throughput_data['count'].sample(sample_buffer_size, replace=True)
            sample_buffer['idx'] = 0

        sample_buffer['idx'] += 1
        return sample_buffer['buffer'].iloc[sample_buffer['idx'] - 1]

    series = {}
    for t in range(trials):
        current_date = start_date
        current_value = start_value

        dates = [current_date]
        steps = [current_value]

        while current_value < target_value:
            current_date += frequency
            current_value += get_sample()

            dates.append(current_date)
            steps.append(min(current_value, target_value))  # don't overshoot the target

        series["Trial %d" % t] = pd.Series(steps, index=dates, name="Trial %d" % t)

    return pd.DataFrame(series)



import logging
import dateutil.parser

import pandas as pd
import matplotlib.pyplot as plt

from ..calculator import Calculator
from ..utils import breakdown_by_month, set_chart_style

logger = logging.getLogger(__name__)

class DefectsCalculator(Calculator):
    """Calculate defect concentration

    Queries JIRA with JQL set in `defects_query` and creates three stacked
    bar charts, presuming their file name values are set. Each shows the
    concentration of defects by month. The number of months to show can be
    limited with `defects_window`.

    - `defects_by_priority_chart`: Grouped by priority
      (`defects_priority_field`), optionally limited to a list of known values
      in order (`defects_priority_values`) and with title
      `defects_by_priority_chart_title`.
    - `defects_by_type_chart`: Grouped by type
      (`defects_type_field`), optionally limited to a list of known values
      in order (`defects_type_values`) and with title
      `defects_by_type_chart_title`.
    - `defects_by_environment_chart`: Grouped by environment
      (`defects_environment_field`), optionally limited to a list of known values
      in order (`defects_environment_values`) and with title
      `defects_by_environment_chart_title`.
    """

    def run(self):

        query = self.settings['defects_query']

        # This calculation is expensive. Only run it if we have a query.
        if not query:
            logger.debug("Not calculating defects chart data as no query specified")
            return None
        
        # Get the fields
        priority_field = self.settings['defects_priority_field']
        priority_field_id = self.query_manager.field_name_to_id(priority_field) if priority_field else None

        type_field = self.settings['defects_type_field']
        type_field_id = self.query_manager.field_name_to_id(type_field) if type_field else None

        environment_field = self.settings['defects_environment_field']
        environment_field_id = self.query_manager.field_name_to_id(environment_field) if environment_field else None
        
        # Build data frame
        columns = ['key', 'priority', 'type', 'environment', 'created', 'resolved']
        series = {
            'key': {'data': [], 'dtype': 'str'},
            'priority': {'data': [], 'dtype': 'str'},
            'type': {'data': [], 'dtype': 'str'},
            'environment': {'data': [], 'dtype': 'str'},
            'created': {'data': [], 'dtype': 'datetime64[ns]'},
            'resolved': {'data': [], 'dtype': 'datetime64[ns]'},
        }

        for issue in self.query_manager.find_issues(query, expand=None):
            series['key']['data'].append(issue.key)
            series['priority']['data'].append(self.query_manager.resolve_field_value(issue, priority_field_id) if priority_field else None)
            series['type']['data'].append(self.query_manager.resolve_field_value(issue, type_field_id) if type_field else None)
            series['environment']['data'].append(self.query_manager.resolve_field_value(issue, environment_field_id) if environment_field else None)
            series['created']['data'].append(dateutil.parser.parse(issue.fields.created))
            series['resolved']['data'].append(dateutil.parser.parse(issue.fields.resolutiondate) if issue.fields.resolutiondate else None)

        data = {}
        for k, v in series.items():
            data[k] = pd.Series(v['data'], dtype=v['dtype'])

        return pd.DataFrame(data, columns=columns)

    def write(self):
        chart_data = self.get_result()
        if chart_data is None:
            return

        if len(chart_data.index) == 0:
            logger.warning("Cannot draw defect charts with zero items")
            return
        
        if self.settings['defects_by_priority_chart']:
            self.write_defects_by_priority_chart(chart_data, self.settings['defects_by_priority_chart'])
        
        if self.settings['defects_by_type_chart']:
            self.write_defects_by_type_chart(chart_data, self.settings['defects_by_type_chart'])
        
        if self.settings['defects_by_environment_chart']:
            self.write_defects_by_environment_chart(chart_data, self.settings['defects_by_environment_chart'])
        
    def write_defects_by_priority_chart(self, chart_data, output_file):
        window = self.settings['defects_window']
        priority_values = self.settings['defects_priority_values']

        breakdown = breakdown_by_month(chart_data, 'created', 'resolved', 'key', 'priority', priority_values)
        
        if window:
            breakdown = breakdown[-window:]
        
        if len(breakdown.index) == 0 or len(breakdown.columns) == 0:
            logger.warning("Cannot draw defects by priority chart with zero items")
            return

        fig, ax = plt.subplots()
        
        breakdown.plot.bar(ax=ax, stacked=True)
        
        if self.settings['defects_by_priority_chart_title']:
            ax.set_title(self.settings['defects_by_priority_chart_title'])

        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        ax.set_xlabel("Month", labelpad=20)
        ax.set_ylabel("Number of items", labelpad=10)

        labels = [d.strftime("%b %y") for d in breakdown.index]
        ax.set_xticklabels(labels, rotation=90, size='small')

        set_chart_style()

        # Write file
        logger.info("Writing defects by priority chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)
    
    def write_defects_by_type_chart(self, chart_data, output_file):
        window = self.settings['defects_window']
        type_values = self.settings['defects_type_values']

        breakdown = breakdown_by_month(chart_data, 'created', 'resolved', 'key', 'type', type_values)
        
        if window:
            breakdown = breakdown[-window:]
        
        if len(breakdown.index) == 0 or len(breakdown.columns) == 0:
            logger.warning("Cannot draw defects by type chart with zero items")
            return

        fig, ax = plt.subplots()
        
        breakdown.plot.bar(ax=ax, stacked=True)
        
        if self.settings['defects_by_type_chart_title']:
            ax.set_title(self.settings['defects_by_type_chart_title'])

        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        ax.set_xlabel("Month", labelpad=20)
        ax.set_ylabel("Number of items", labelpad=10)

        labels = [d.strftime("%b %y") for d in breakdown.index]
        ax.set_xticklabels(labels, rotation=90, size='small')

        set_chart_style()

        # Write file
        logger.info("Writing defects by type chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)
    
    def write_defects_by_environment_chart(self, chart_data, output_file):
        window = self.settings['defects_window']
        environment_values = self.settings['defects_environment_values']

        breakdown = breakdown_by_month(chart_data, 'created', 'resolved', 'key', 'environment', environment_values)
        
        if window:
            breakdown = breakdown[-window:]
        
        if len(breakdown.index) == 0 or len(breakdown.columns) == 0:
            logger.warning("Cannot draw defects by environment chart with zero items")
            return

        fig, ax = plt.subplots()
        
        breakdown.plot.bar(ax=ax, stacked=True)
        
        if self.settings['defects_by_environment_chart_title']:
            ax.set_title(self.settings['defects_by_environment_chart_title'])

        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        ax.set_xlabel("Month", labelpad=20)
        ax.set_ylabel("Number of items", labelpad=10)

        labels = [d.strftime("%b %y") for d in breakdown.index]
        ax.set_xticklabels(labels, rotation=90, size='small')

        set_chart_style()

        # Write file
        logger.info("Writing defects by environment chart to %s", output_file)
        fig.savefig(output_file, bbox_inches='tight', dpi=300)
        plt.close(fig)



import pytest
import math
from pandas import DataFrame, Timedelta

from .cycletime import CycleTimeCalculator
from .percentiles import PercentilesCalculator

from ..utils import extend_dict

@pytest.fixture
def settings(minimal_settings):
    return extend_dict(minimal_settings, {
        'quantiles': [0.1, 0.5, 0.9]
    })

@pytest.fixture
def query_manager(minimal_query_manager):
    return minimal_query_manager

@pytest.fixture
def results(large_cycle_time_results):
    return extend_dict(large_cycle_time_results, {})

def test_empty(query_manager, settings, minimal_cycle_time_columns):
    results = {
        CycleTimeCalculator: DataFrame([], columns=minimal_cycle_time_columns, index=[])
    }

    calculator = PercentilesCalculator(query_manager, settings, results)

    data = calculator.run()

    assert list(data.index) == [0.1, 0.5, 0.9]
    assert math.isnan(list(data)[0])
    assert math.isnan(list(data)[1])
    assert math.isnan(list(data)[2])

def test_calculate_percentiles(query_manager, settings, results):
    calculator = PercentilesCalculator(query_manager, settings, results)

    data = calculator.run()

    assert list(data.index) == [0.1, 0.5, 0.9]
    assert list(data) == [Timedelta('4 days 12:00:00'), Timedelta('5 days 00:00:00'), Timedelta('5 days 00:00:00')]



import datetime

from .config import (
    force_list,
    expand_key,
    config_to_options
)

def test_force_list():
    assert force_list(None) == [None]
    assert force_list("foo") == ["foo"]
    assert force_list(("foo",)) == ["foo"]
    assert force_list(["foo"]) == ["foo"]

def test_expand_key():
    assert expand_key("foo") == "foo"
    assert expand_key("foo_bar") == "foo bar"
    assert expand_key("FOO") == "foo"
    assert expand_key("FOO_bar") == "foo bar"

def test_config_to_options_minimal():

    options = config_to_options("""\
Connection:
    Domain: https://foo.com

Query: (filter=123)

Workflow:
    Backlog: Backlog
    In progress: Build
    Done: Done
""")

    assert options['connection']['domain'] == 'https://foo.com'
    assert options['settings']['queries'][0] == {'value': None, 'jql': '(filter=123)'}

    assert options['settings']['backlog_column'] == 'Backlog'
    assert options['settings']['committed_column'] == 'In progress'
    assert options['settings']['final_column'] == 'In progress'
    assert options['settings']['done_column'] == 'Done'
    

def test_config_to_options_maximal():

    options = config_to_options("""\
Connection:
    Domain: https://foo.com
    Username: user1
    Password: apassword
    HTTP Proxy: https://proxy1.local
    HTTPS Proxy: https://proxy2.local

Queries:
    Attribute: Team
    Criteria:
        - Value: Team 1
          JQL: (filter=123)

        - Value: Team 2
          JQL: (filter=124)

Attributes:
    Team: Team
    Release: Fix version/s

Known values:
    Release:
        - "R01"
        - "R02"
        - "R03"

Workflow:
    Backlog: Backlog
    Committed: Next
    Build: Build
    Test:
        - Code review
        - QA
    Done: Done

Output:
    Quantiles:
        - 0.1
        - 0.2

    Backlog column: Backlog
    Committed column: Committed
    Final column: Test
    Done column: Done

    Cycle time data: cycletime.csv
    Percentiles data: percentiles.csv
    
    Scatterplot window: 30
    Scatterplot data: scatterplot.csv
    Scatterplot chart: scatterplot.png
    Scatterplot chart title: Cycle time scatter plot
    
    Histogram window: 30
    Histogram chart: histogram.png
    Histogram chart title: Cycle time histogram

    CFD window: 30
    CFD data: cfd.csv
    CFD chart: cfd.png
    CFD chart title: Cumulative Flow Diagram

    Histogram data: histogram.csv
    
    Throughput frequency: 1D
    Throughput window: 3
    Throughput data: throughput.csv
    Throughput chart: throughput.png
    Throughput chart title: Throughput trend

    Burnup window: 30
    Burnup chart: burnup.png
    Burnup chart title: Burn-up

    Burnup forecast window: 30
    Burnup forecast chart: burnup-forecast.png
    Burnup forecast chart title: Burn-up forecast
    Burnup forecast chart target: 100
    Burnup forecast chart deadline: 2018-06-01
    Burnup forecast chart deadline confidence: .85
    Burnup forecast chart trials: 50
    Burnup forecast chart throughput window: 30
    Burnup forecast chart throughput window end: 2018-03-01

    WIP frequency: 3D
    WIP window: 3
    WIP chart: wip.png
    WIP chart title: Work in Progress

    Ageing WIP chart: ageing-wip.png
    Ageing WIP chart title: Ageing WIP

    Net flow frequency: 5D
    Net flow window: 3
    Net flow chart: net-flow.png
    Net flow chart title: Net flow

    Impediments data: impediments.csv
    Impediments window: 3
    Impediments chart: impediments.png
    Impediments chart title: Impediments
    Impediments days chart: impediments-days.png
    Impediments days chart title: Total impeded days
    Impediments status chart: impediments-status.png
    Impediments status chart title: Impediments by status
    Impediments status days chart: impediments-status-days.png
    Impediments status days chart title: Total impeded days by status

    Defects query: issueType = Bug
    Defects window: 3
    Defects priority field: Priority
    Defects priority values:
        - Low
        - Medium
        - High
    Defects type field: Type
    Defects type values:
        - Config
        - Data
        - Code
    Defects environment field: Environment
    Defects environment values:
        - SIT
        - UAT
        - PROD
    Defects by priority chart: defects-by-priority.png
    Defects by priority chart title: Defects by priority
    Defects by type chart: defects-by-type.png
    Defects by type chart title: Defects by type
    Defects by environment chart: defects-by-environment.png
    Defects by environment chart title: Defects by environment
    
    Debt query: issueType = "Tech debt"
    Debt window: 3
    Debt priority field: Priority
    Debt priority values:
        - Low
        - Medium
        - High
    Debt chart: tech-debt.png
    Debt chart title: Technical debt
    Debt age chart: tech-debt-age.png
    Debt age chart title: Technical debt age
    Debt age chart bins:
        - 10
        - 20
        - 30

    Waste query: issueType = Story AND resolution IN (Withdrawn, Invalid)
    Waste window: 3
    Waste frequency: 2W-WED
    Waste chart: waste.png
    Waste chart title: Waste
""")

    assert options['connection'] == {
        'domain': 'https://foo.com',
        'jira_client_options': {},
        'password': 'apassword',
        'username': 'user1',
        'http_proxy': 'https://proxy1.local',
        'https_proxy': 'https://proxy2.local',
    }
   
    assert options['settings'] == {
        'cycle': [
            {'name': 'Backlog', 'statuses': ['Backlog'], 'type': 'backlog'},
            {'name': 'Committed', 'statuses': ['Next'], 'type': 'accepted'},
            {'name': 'Build', 'statuses': ['Build'], 'type': 'accepted'},
            {'name': 'Test',
                'statuses': ['Code review', 'QA'],
                'type': 'accepted'},
            {'name': 'Done', 'statuses': ['Done'], 'type': 'complete'}
        ],

        'attributes': {'Release': 'Fix version/s', 'Team': 'Team'},
        'known_values': {'Release': ['R01', 'R02', 'R03']},
        'max_results': None,
        'verbose': False,

        'queries': [{'jql': '(filter=123)', 'value': 'Team 1'},
                    {'jql': '(filter=124)', 'value': 'Team 2'}],
        'query_attribute': 'Team',
                
        'backlog_column': 'Backlog',
        'committed_column': 'Committed',
        'final_column': 'Test',
        'done_column': 'Done',

        'quantiles': [0.1, 0.2],

        'cycle_time_data': 'cycletime.csv',
        
        'ageing_wip_chart': 'ageing-wip.png',
        'ageing_wip_chart_title': 'Ageing WIP',
        
        'burnup_window': 30,
        'burnup_chart': 'burnup.png',
        'burnup_chart_title': 'Burn-up',
        
        'burnup_forecast_window': 30,
        'burnup_forecast_chart': 'burnup-forecast.png',
        'burnup_forecast_chart_deadline': datetime.date(2018, 6, 1),
        'burnup_forecast_chart_deadline_confidence': 0.85,
        'burnup_forecast_chart_target': 100,
        'burnup_forecast_chart_throughput_window': 30,
        'burnup_forecast_chart_throughput_window_end': datetime.date(2018, 3, 1),
        'burnup_forecast_chart_title': 'Burn-up forecast',
        'burnup_forecast_chart_trials': 50,
        
        'cfd_window': 30,
        'cfd_chart': 'cfd.png',
        'cfd_chart_title': 'Cumulative Flow Diagram',
        'cfd_data': 'cfd.csv',
        
        'histogram_window': 30,
        'histogram_chart': 'histogram.png',
        'histogram_chart_title': 'Cycle time histogram',
        'histogram_data': 'histogram.csv',
        
        'net_flow_frequency': '5D',
        'net_flow_window': 3,
        'net_flow_chart': 'net-flow.png',
        'net_flow_chart_title': 'Net flow',
        
        'percentiles_data': 'percentiles.csv',
        
        'scatterplot_window': 30,
        'scatterplot_chart': 'scatterplot.png',
        'scatterplot_chart_title': 'Cycle time scatter plot',
        'scatterplot_data': 'scatterplot.csv',
        
        'throughput_frequency': '1D',
        'throughput_window': 3,
        'throughput_chart': 'throughput.png',
        'throughput_chart_title': 'Throughput trend',
        'throughput_data': 'throughput.csv',
        
        'wip_frequency': '3D',
        'wip_window': 3,
        'wip_chart': 'wip.png',
        'wip_chart_title': 'Work in Progress',

        'impediments_data': 'impediments.csv',
        'impediments_window': 3,
        'impediments_chart': 'impediments.png',
        'impediments_chart_title': 'Impediments',
        'impediments_days_chart': 'impediments-days.png',
        'impediments_days_chart_title': 'Total impeded days',
        'impediments_status_chart': 'impediments-status.png',
        'impediments_status_chart_title': 'Impediments by status',
        'impediments_status_days_chart': 'impediments-status-days.png',
        'impediments_status_days_chart_title': 'Total impeded days by status',

        'defects_query': 'issueType = Bug',
        'defects_window': 3,
        'defects_priority_field': 'Priority',
        'defects_priority_values': ['Low', 'Medium', 'High'],
        'defects_type_field': 'Type',
        'defects_type_values': ['Config', 'Data', 'Code'],
        'defects_environment_field': 'Environment',
        'defects_environment_values': ['SIT', 'UAT', 'PROD'],

        'defects_by_priority_chart': 'defects-by-priority.png',
        'defects_by_priority_chart_title': 'Defects by priority',
        'defects_by_type_chart': 'defects-by-type.png',
        'defects_by_type_chart_title': 'Defects by type',
        'defects_by_environment_chart': 'defects-by-environment.png',
        'defects_by_environment_chart_title': 'Defects by environment',
        
        'debt_query': 'issueType = "Tech debt"',
        'debt_window': 3,
        'debt_priority_field': 'Priority',
        'debt_priority_values': ['Low', 'Medium', 'High'],
        'debt_chart': 'tech-debt.png',
        'debt_chart_title': 'Technical debt',
        'debt_age_chart': 'tech-debt-age.png',
        'debt_age_chart_title': 'Technical debt age',
        'debt_age_chart_bins': [10, 20, 30],

        'waste_query': 'issueType = Story AND resolution IN (Withdrawn, Invalid)',
        'waste_window': 3,
        'waste_frequency': '2W-WED',
        'waste_chart': 'waste.png',
        'waste_chart_title': 'Waste',
        
    }

def test_config_to_options_strips_directories():

    options = config_to_options("""\
Connection:
    Domain: https://foo.com

Query: (filter=123)

Workflow:
    Backlog: Backlog
    In progress: Build
    Done: Done

Output:
    Cycle time data: tmp/cycletime.csv
    Percentiles data: /tmp/percentiles.csv
    Scatterplot data: ../scatterplot.csv
    Scatterplot chart: /foo/bar/baz/tmp/scatterplot.png
    Histogram chart: tmp/histogram.png
    CFD data: tmp/cfd.csv
    CFD chart: tmp/cfd.png
    Histogram data: tmp/histogram.csv
    Throughput data: tmp/throughput.csv
    Throughput chart: tmp/throughput.png
    Burnup chart: tmp/burnup.png
    Burnup forecast chart: tmp/burnup-forecast.png
    WIP chart: tmp/wip.png
    Ageing WIP chart: tmp/ageing-wip.png
    Net flow chart: tmp/net-flow.png
""")

    assert options['settings']['cycle_time_data'] == 'cycletime.csv'
    assert options['settings']['ageing_wip_chart'] == 'ageing-wip.png'
    assert options['settings']['burnup_chart'] == 'burnup.png'
    assert options['settings']['burnup_forecast_chart'] == 'burnup-forecast.png'
    assert options['settings']['cfd_chart'] == 'cfd.png'
    assert options['settings']['histogram_chart'] == 'histogram.png'
    assert options['settings']['histogram_data'] == 'histogram.csv'
    assert options['settings']['net_flow_chart'] == 'net-flow.png'
    assert options['settings']['percentiles_data'] == 'percentiles.csv'
    assert options['settings']['scatterplot_chart'] == 'scatterplot.png'
    assert options['settings']['scatterplot_data'] == 'scatterplot.csv'
    assert options['settings']['throughput_chart'] == 'throughput.png'
    assert options['settings']['throughput_data'] == 'throughput.csv'
    assert options['settings']['wip_chart'] == 'wip.png'



import pytest
import datetime
from pandas import NaT, Timestamp, Timedelta

from ..conftest import (
    FauxJIRA as JIRA,
    FauxIssue as Issue,
    FauxChange as Change,
    FauxFieldValue as Value
)

from ..querymanager import QueryManager
from .cycletime import CycleTimeCalculator

@pytest.fixture
def jira(custom_fields):
    return JIRA(fields=custom_fields, issues=[
        Issue("A-1",
            summary="Just created",
            issuetype=Value("Story", "story"),
            status=Value("Backlog", "backlog"),
            resolution=None,
            resolutiondate=None,
            created="2018-01-01 01:01:01",
            customfield_001="Team 1",
            customfield_002=Value(None, 10),
            customfield_003=Value(None, ["R2", "R3", "R4"]),
            customfield_100=None,
            changes=[],
        ),
        Issue("A-2",
            summary="Started",
            issuetype=Value("Story", "story"),
            status=Value("Next", "next"),
            resolution=None,
            resolutiondate=None,
            created="2018-01-02 01:01:01",
            customfield_001="Team 1",
            customfield_002=Value(None, 20),
            customfield_003=Value(None, []),
            customfield_100=None,
            changes=[
                Change("2018-01-02 10:01:01", [("Flagged", None, "Impediment")]),
                Change("2018-01-03 01:00:00", [("Flagged", "Impediment", "")]),  # blocked 1 day in the backlog (doesn't count towards blocked days)
                Change("2018-01-03 01:01:01", [("status", "Backlog", "Next",)]),
                Change("2018-01-04 10:01:01", [("Flagged", "", "Impediment")]),
                Change("2018-01-05 08:01:01", [("Flagged", "Impediment", "")]),  # was blocked 1 day
                Change("2018-01-08 10:01:01", [("Flagged", "", "Impediment")]),  # stays blocked until today
            ],
        ),
        Issue("A-3",
            summary="Completed",
            issuetype=Value("Story", "story"),
            status=Value("Done", "done"),
            resolution=Value("Done", "Done"),
            resolutiondate="2018-01-06 01:01:01",
            created="2018-01-03 01:01:01",
            customfield_001="Team 1",
            customfield_002=Value(None, 30),
            customfield_003=Value(None, []),
            customfield_100=None,
            changes=[
                Change("2018-01-03 01:01:01", [("status", "Backlog", "Next",)]),
                Change("2018-01-04 01:01:01", [("status", "Next", "Build",)]),
                Change("2018-01-04 10:01:01", [("Flagged", None, "Impediment")]),  # should clear two days later when issue resolved
                Change("2018-01-05 01:01:01", [("status", "Build", "QA",)]),
                Change("2018-01-06 01:01:01", [("status", "QA", "Done",)]),
            ],
        ),
        Issue("A-4",
            summary="Moved back",
            issuetype=Value("Story", "story"),
            status=Value("Next", "next"),
            resolution=None,
            resolutiondate=None,
            created="2018-01-04 01:01:01",
            customfield_001="Team 1",
            customfield_002=Value(None, 30),
            customfield_003=Value(None, []),
            customfield_100=None,
            changes=[
                Change("2018-01-04 01:01:01", [("status", "Backlog", "Next",)]),
                Change("2018-01-05 01:01:01", [("status", "Next", "Build",)]),
                Change("2018-01-06 01:01:01", [("status", "Build", "Next",)]),
                Change("2018-01-07 01:01:01", [("Flagged", None, "Awaiting input")]),
                Change("2018-01-10 10:01:01", [("Flagged", "Awaiting input", "")]),  # blocked 3 days
            ],
        ),
    ])

@pytest.fixture
def settings(custom_settings):
    return custom_settings

def test_columns(jira, settings):
    query_manager = QueryManager(jira, settings)
    results = {}
    calculator = CycleTimeCalculator(query_manager, settings, results)

    data = calculator.run()

    assert list(data.columns) == [
        'key',
        'url',
        'issue_type',
        'summary',
        'status',
        'resolution',

        'Estimate',
        'Release',
        'Team',
        
        'cycle_time',
        'completed_timestamp',
        'blocked_days',
        'impediments',
        
        'Backlog',
        'Committed',
        'Build',
        'Test',
        'Done'
    ]

def test_empty(custom_fields, settings):
    jira = JIRA(fields=custom_fields, issues=[])
    query_manager = QueryManager(jira, settings)
    results = {}
    calculator = CycleTimeCalculator(query_manager, settings, results)

    data = calculator.run()

    assert len(data.index) == 0

def test_movement(jira, settings):
    query_manager = QueryManager(jira, settings)
    results = {}
    calculator = CycleTimeCalculator(query_manager, settings, results)

    data = calculator.run(now=datetime.datetime(2018, 1, 10, 15, 37, 0))

    assert data.to_dict('records') == [{
        'key': 'A-1',
        'url': 'https://example.org/browse/A-1',
        'issue_type': 'Story',
        'summary': 'Just created',
        'status': 'Backlog',
        'resolution': None,

        'Estimate': 10,
        'Release': 'R3',
        'Team': 'Team 1',

        'completed_timestamp': NaT,
        'cycle_time': NaT,
        'blocked_days': 0,
        'impediments': [],

        'Backlog': Timestamp('2018-01-01 01:01:01'),
        'Committed': NaT,
        'Build': NaT,
        'Test': NaT,
        'Done': NaT,
    }, {
        'key': 'A-2',
        'url': 'https://example.org/browse/A-2',
        'issue_type': 'Story',
        'summary': 'Started',
        'status': 'Next',
        'resolution': None,

        'Estimate': 20,
        'Release': 'None',
        'Team': 'Team 1',

        'completed_timestamp': NaT,
        'cycle_time': NaT,
        'blocked_days': 3,
        'impediments': [
            {'start': datetime.date(2018, 1, 2), 'end': datetime.date(2018, 1, 3), 'status': 'Backlog', 'flag': 'Impediment'},  # doesn't count towards blocked_days
            {'start': datetime.date(2018, 1, 4), 'end': datetime.date(2018, 1, 5), 'status': 'Committed', 'flag': 'Impediment'},
            {'start': datetime.date(2018, 1, 8), 'end': None, 'status': 'Committed', 'flag': 'Impediment'},
        ],

        'Backlog': Timestamp('2018-01-02 01:01:01'),
        'Committed': Timestamp('2018-01-03 01:01:01'),
        'Build': NaT,
        'Test': NaT,
        'Done': NaT,
    }, {
        'key': 'A-3',
        'url': 'https://example.org/browse/A-3',
        'summary': 'Completed',
        'issue_type': 'Story',
        'status': 'Done',
        'resolution': 'Done',

        'Estimate': 30,
        'Release': 'None',
        'Team': 'Team 1',

        'completed_timestamp': Timestamp('2018-01-06 01:01:01'),
        'cycle_time': Timedelta('3 days 00:00:00'),
        'blocked_days': 2,
        'impediments': [{'start': datetime.date(2018, 1, 4), 'end': datetime.date(2018, 1, 6), 'status': 'Build', 'flag': 'Impediment'}],

        'Backlog': Timestamp('2018-01-03 01:01:01'),
        'Committed': Timestamp('2018-01-03 01:01:01'),
        'Build': Timestamp('2018-01-04 01:01:01'),
        'Test': Timestamp('2018-01-05 01:01:01'),
        'Done': Timestamp('2018-01-06 01:01:01'),
    }, {
        'key': 'A-4',
        'url': 'https://example.org/browse/A-4',
        'summary': 'Moved back',
        'issue_type': 'Story',
        'status': 'Next',
        'resolution': None,

        'Estimate': 30,
        'Release': 'None',
        'Team': 'Team 1',
        
        'completed_timestamp': NaT,
        'cycle_time': NaT,
        'blocked_days': 3,
        'impediments': [{'start': datetime.date(2018, 1, 7), 'end': datetime.date(2018, 1, 10), 'status': 'Committed', 'flag': 'Awaiting input'}],

        'Backlog': Timestamp('2018-01-04 01:01:01'),
        'Committed': Timestamp('2018-01-04 01:01:01'),
        'Build': NaT,
        'Test': NaT,
        'Done': NaT,
    }]



import datetime
import numpy as np
import pandas as pd

from .utils import (
    get_extension,
    to_json_string,
    to_days_since_epoch,
    extend_dict,
    breakdown_by_month,
    breakdown_by_month_sum_days,
    to_bin
)

def test_extend_dict():
    assert extend_dict({'one': 1}, {'two': 2}) == {'one': 1, 'two': 2}

def test_get_extension():
    assert get_extension("foo.csv") == ".csv"
    assert get_extension("/path/to/foo.csv") == ".csv"
    assert get_extension("\\path\\to\\foo.csv") == ".csv"
    assert get_extension("foo") == ""
    assert get_extension("foo.CSV") == ".csv"

def test_to_json_string():
    assert to_json_string(1) == "1"
    assert to_json_string("foo") == "foo"
    assert to_json_string(None) == ""
    assert to_json_string(np.NaN) == ""
    assert to_json_string(pd.NaT) == ""
    assert to_json_string(pd.Timestamp(2018, 2, 1)) == "2018-02-01"

def test_to_days_since_epoch():
    assert to_days_since_epoch(datetime.date(1970, 1, 1)) == 0
    assert to_days_since_epoch(datetime.date(1970, 1, 15)) == 14

def test_breakdown_by_month():

    df = pd.DataFrame([
        {'key': 'ABC-1', 'priority': 'high', 'start': pd.Timestamp(2018, 1, 1), 'end': pd.Timestamp(2018, 3, 20)},
        {'key': 'ABC-2', 'priority': 'med',  'start': pd.Timestamp(2018, 1, 2), 'end': pd.Timestamp(2018, 1, 20)},
        {'key': 'ABC-3', 'priority': 'high', 'start': pd.Timestamp(2018, 2, 3), 'end': pd.Timestamp(2018, 3, 20)},
        {'key': 'ABC-4', 'priority': 'med',  'start': pd.Timestamp(2018, 1, 4), 'end': pd.Timestamp(2018, 3, 20)},
        {'key': 'ABC-5', 'priority': 'high', 'start': pd.Timestamp(2018, 2, 5), 'end': pd.Timestamp(2018, 2, 20)},
        {'key': 'ABC-6', 'priority': 'med',  'start': pd.Timestamp(2018, 3, 6), 'end': pd.Timestamp(2018, 3, 20)}
    ], columns=['key', 'priority', 'start', 'end'])

    breakdown = breakdown_by_month(df, 'start', 'end', 'key', 'priority', ['low', 'med', 'high'])
    assert list(breakdown.columns) == ['med', 'high']
    
    assert list(breakdown.index) == [
        pd.Timestamp(2018, 1, 1),
        pd.Timestamp(2018, 2, 1),
        pd.Timestamp(2018, 3, 1),
    ]
    assert breakdown.to_dict('records') == [
        {'high': 1, 'med': 2},
        {'high': 3, 'med': 1},
        {'high': 2, 'med': 2},
    ]

def test_breakdown_by_month_open_ended():

    df = pd.DataFrame([
        {'key': 'ABC-1', 'priority': 'high', 'start': pd.Timestamp(2018, 1, 1), 'end': pd.Timestamp(2018, 3, 20)},
        {'key': 'ABC-2', 'priority': 'med',  'start': pd.Timestamp(2018, 1, 2), 'end': pd.Timestamp(2018, 1, 20)},
        {'key': 'ABC-3', 'priority': 'high', 'start': pd.Timestamp(2018, 2, 3), 'end': pd.Timestamp(2018, 3, 20)},
        {'key': 'ABC-4', 'priority': 'med',  'start': pd.Timestamp(2018, 1, 4), 'end': pd.Timestamp(2018, 3, 20)},
        {'key': 'ABC-5', 'priority': 'high', 'start': pd.Timestamp(2018, 2, 5), 'end': pd.Timestamp(2018, 2, 20)},
        {'key': 'ABC-6', 'priority': 'med',  'start': pd.Timestamp(2018, 3, 6), 'end': None}
    ], columns=['key', 'priority', 'start', 'end'])

    breakdown = breakdown_by_month(df, 'start', 'end', 'key', 'priority', ['low', 'med', 'high'])
    assert list(breakdown.columns) == ['med', 'high']

    # Note: We will get columns until the current month; assume this test is
    # run from June onwards ;)
    
    assert list(breakdown.index)[:5] == [
        pd.Timestamp(2018, 1, 1),
        pd.Timestamp(2018, 2, 1),
        pd.Timestamp(2018, 3, 1),
        pd.Timestamp(2018, 4, 1),
        pd.Timestamp(2018, 5, 1),
    ]
    assert breakdown.to_dict('records')[:5] == [
        {'high': 1, 'med': 2},
        {'high': 3, 'med': 1},
        {'high': 2, 'med': 2},
        {'high': 0, 'med': 1},
        {'high': 0, 'med': 1},
    ]

def test_breakdown_by_month_no_column_spec():

    df = pd.DataFrame([
        {'key': 'ABC-1', 'priority': 'high', 'start': pd.Timestamp(2018, 1, 1), 'end': pd.Timestamp(2018, 3, 20)},
        {'key': 'ABC-2', 'priority': 'med',  'start': pd.Timestamp(2018, 1, 2), 'end': pd.Timestamp(2018, 1, 20)},
        {'key': 'ABC-3', 'priority': 'high', 'start': pd.Timestamp(2018, 2, 3), 'end': pd.Timestamp(2018, 3, 20)},
        {'key': 'ABC-4', 'priority': 'med',  'start': pd.Timestamp(2018, 1, 4), 'end': pd.Timestamp(2018, 3, 20)},
        {'key': 'ABC-5', 'priority': 'high', 'start': pd.Timestamp(2018, 2, 5), 'end': pd.Timestamp(2018, 2, 20)},
        {'key': 'ABC-6', 'priority': 'med',  'start': pd.Timestamp(2018, 3, 6), 'end': pd.Timestamp(2018, 3, 20)}
    ], columns=['key', 'priority', 'start', 'end'])

    breakdown = breakdown_by_month(df, 'start', 'end', 'key', 'priority')
    assert list(breakdown.columns) == ['high', 'med']  # alphabetical
    
    assert list(breakdown.index) == [
        pd.Timestamp(2018, 1, 1),
        pd.Timestamp(2018, 2, 1),
        pd.Timestamp(2018, 3, 1),
    ]
    assert breakdown.to_dict('records') == [
        {'high': 1, 'med': 2},
        {'high': 3, 'med': 1},
        {'high': 2, 'med': 2},
    ]

def test_breakdown_by_month_none_values():

    df = pd.DataFrame([
        {'key': 'ABC-1', 'priority': None, 'start': pd.Timestamp(2018, 1, 1), 'end': pd.Timestamp(2018, 3, 20)},
        {'key': 'ABC-2', 'priority': None, 'start': pd.Timestamp(2018, 1, 2), 'end': pd.Timestamp(2018, 1, 20)},
        {'key': 'ABC-3', 'priority': None, 'start': pd.Timestamp(2018, 2, 3), 'end': pd.Timestamp(2018, 3, 20)},
        {'key': 'ABC-4', 'priority': None, 'start': pd.Timestamp(2018, 1, 4), 'end': pd.Timestamp(2018, 3, 20)},
        {'key': 'ABC-5', 'priority': None, 'start': pd.Timestamp(2018, 2, 5), 'end': pd.Timestamp(2018, 2, 20)},
        {'key': 'ABC-6', 'priority': None, 'start': pd.Timestamp(2018, 3, 6), 'end': pd.Timestamp(2018, 3, 20)}
    ], columns=['key', 'priority', 'start', 'end'])

    breakdown = breakdown_by_month(df, 'start', 'end', 'key', 'priority')
    assert list(breakdown.columns) == [None]
    
    assert list(breakdown.index) == [
        pd.Timestamp(2018, 1, 1),
        pd.Timestamp(2018, 2, 1),
        pd.Timestamp(2018, 3, 1),
    ]
    assert breakdown.to_dict('records') == [{None: 3}, {None: 4}, {None: 4}]

def test_breakdown_by_month_sum_days():

    df = pd.DataFrame([
        {'priority': 'high', 'start': pd.Timestamp(2018, 1, 1), 'end': pd.Timestamp(2018, 3, 20)},  # Jan: 31 Feb: 28 Mar: 20
        {'priority': 'med',  'start': pd.Timestamp(2018, 1, 2), 'end': pd.Timestamp(2018, 1, 20)},  # Jan: 19 Feb:  0 Mar:  0
        {'priority': 'high', 'start': pd.Timestamp(2018, 2, 3), 'end': pd.Timestamp(2018, 3, 20)},  # Jan:  0 Feb: 26 Mar: 20
        {'priority': 'med',  'start': pd.Timestamp(2018, 1, 4), 'end': pd.Timestamp(2018, 3, 20)},  # Jan: 28 Feb: 28 Mar: 20
        {'priority': 'high', 'start': pd.Timestamp(2018, 2, 5), 'end': pd.Timestamp(2018, 2, 20)},  # Jan:  0 Feb: 16 Mar:  0
        {'priority': 'med',  'start': pd.Timestamp(2018, 3, 6), 'end': pd.Timestamp(2018, 3, 20)}   # Jan:  0 Feb:  0 Mar: 15
    ], columns=['key', 'priority', 'start', 'end'])

    breakdown = breakdown_by_month_sum_days(df, 'start', 'end', 'priority', ['low', 'med', 'high'])
    assert list(breakdown.columns) == ['med', 'high']
    
    assert list(breakdown.index) == [
        pd.Timestamp(2018, 1, 1),
        pd.Timestamp(2018, 2, 1),
        pd.Timestamp(2018, 3, 1),
    ]

    assert breakdown.to_dict('records') == [
        {'high': 31.0, 'med': 47.0},
        {'high': 70.0, 'med': 28.0},
        {'high': 40.0, 'med': 35.0}
    ]

def test_breakdown_by_month_sum_days_no_column_spec():

    df = pd.DataFrame([
        {'priority': 'high', 'start': pd.Timestamp(2018, 1, 1), 'end': pd.Timestamp(2018, 3, 20)},  # Jan: 31 Feb: 28 Mar: 20
        {'priority': 'med',  'start': pd.Timestamp(2018, 1, 2), 'end': pd.Timestamp(2018, 1, 20)},  # Jan: 19 Feb:  0 Mar:  0
        {'priority': 'high', 'start': pd.Timestamp(2018, 2, 3), 'end': pd.Timestamp(2018, 3, 20)},  # Jan:  0 Feb: 26 Mar: 20
        {'priority': 'med',  'start': pd.Timestamp(2018, 1, 4), 'end': pd.Timestamp(2018, 3, 20)},  # Jan: 28 Feb: 28 Mar: 20
        {'priority': 'high', 'start': pd.Timestamp(2018, 2, 5), 'end': pd.Timestamp(2018, 2, 20)},  # Jan:  0 Feb: 16 Mar:  0
        {'priority': 'med',  'start': pd.Timestamp(2018, 3, 6), 'end': pd.Timestamp(2018, 3, 20)}   # Jan:  0 Feb:  0 Mar: 15
    ], columns=['key', 'priority', 'start', 'end'])

    breakdown = breakdown_by_month_sum_days(df, 'start', 'end', 'priority')
    assert list(breakdown.columns) == ['high', 'med']  # alphabetical
    
    assert list(breakdown.index) == [
        pd.Timestamp(2018, 1, 1),
        pd.Timestamp(2018, 2, 1),
        pd.Timestamp(2018, 3, 1),
    ]

    assert breakdown.to_dict('records') == [
        {'high': 31.0, 'med': 47.0},
        {'high': 70.0, 'med': 28.0},
        {'high': 40.0, 'med': 35.0}
    ]

def test_breakdown_by_month_sum_day_open_ended():

    df = pd.DataFrame([
        {'priority': 'high', 'start': pd.Timestamp(2018, 1, 1), 'end': pd.Timestamp(2018, 3, 20)},  # Jan: 31 Feb: 28 Mar: 20
        {'priority': 'med',  'start': pd.Timestamp(2018, 1, 2), 'end': pd.Timestamp(2018, 1, 20)},  # Jan: 19 Feb:  0 Mar:  0
        {'priority': 'high', 'start': pd.Timestamp(2018, 2, 3), 'end': pd.Timestamp(2018, 3, 20)},  # Jan:  0 Feb: 26 Mar: 20
        {'priority': 'med',  'start': pd.Timestamp(2018, 1, 4), 'end': pd.Timestamp(2018, 3, 20)},  # Jan: 28 Feb: 28 Mar: 20
        {'priority': 'high', 'start': pd.Timestamp(2018, 2, 5), 'end': pd.Timestamp(2018, 2, 20)},  # Jan:  0 Feb: 16 Mar:  0
        {'priority': 'med',  'start': pd.Timestamp(2018, 3, 6), 'end': None}                        # Jan:  0 Feb:  0 Mar: 26
    ], columns=['key', 'priority', 'start', 'end'])

    breakdown = breakdown_by_month_sum_days(df, 'start', 'end', 'priority', ['low', 'med', 'high'])
    assert list(breakdown.columns) == ['med', 'high']

    # Note: We will get columns until the current month; assume this test is
    # run from June onwards ;)
    
    assert list(breakdown.index)[:5] == [
        pd.Timestamp(2018, 1, 1),
        pd.Timestamp(2018, 2, 1),
        pd.Timestamp(2018, 3, 1),
        pd.Timestamp(2018, 4, 1),
        pd.Timestamp(2018, 5, 1),
    ]
    assert breakdown.to_dict('records')[:5] == [
        {'high': 31.0, 'med': 47.0},
        {'high': 70.0, 'med': 28.0},
        {'high': 40.0, 'med': 46.0},
        {'high': 0,    'med': 30.0},
        {'high': 0,    'med': 31.0},
    ]

def test_breakdown_by_month_sum_days_none_values():

    df = pd.DataFrame([
        {'priority': None,   'start': pd.Timestamp(2018, 1, 1), 'end': pd.Timestamp(2018, 3, 20)},  # Jan: 31 Feb: 28 Mar: 20
        {'priority': 'med',  'start': pd.Timestamp(2018, 1, 2), 'end': pd.Timestamp(2018, 1, 20)},  # Jan: 19 Feb:  0 Mar:  0
        {'priority': None,   'start': pd.Timestamp(2018, 2, 3), 'end': pd.Timestamp(2018, 3, 20)},  # Jan:  0 Feb: 26 Mar: 20
        {'priority': 'med',  'start': pd.Timestamp(2018, 1, 4), 'end': pd.Timestamp(2018, 3, 20)},  # Jan: 28 Feb: 28 Mar: 20
        {'priority': None,   'start': pd.Timestamp(2018, 2, 5), 'end': pd.Timestamp(2018, 2, 20)},  # Jan:  0 Feb: 16 Mar:  0
        {'priority': 'med',  'start': pd.Timestamp(2018, 3, 6), 'end': pd.Timestamp(2018, 3, 20)}   # Jan:  0 Feb:  0 Mar: 15
    ], columns=['key', 'priority', 'start', 'end'])

    breakdown = breakdown_by_month_sum_days(df, 'start', 'end', 'priority')
    assert list(breakdown.columns) == [None, 'med']
    
    assert list(breakdown.index) == [
        pd.Timestamp(2018, 1, 1),
        pd.Timestamp(2018, 2, 1),
        pd.Timestamp(2018, 3, 1),
    ]

    assert breakdown.to_dict('records') == [
        {None: 31.0, 'med': 47.0},
        {None: 70.0, 'med': 28.0},
        {None: 40.0, 'med': 35.0}
    ]

def test_to_bin():

    assert to_bin(0, [10, 20, 30]) == (0, 10)
    assert to_bin(9, [10, 20, 30]) == (0, 10)
    assert to_bin(10, [10, 20, 30]) == (0, 10)

    assert to_bin(11, [10, 20, 30]) == (10, 20)
    assert to_bin(20, [10, 20, 30]) == (10, 20)

    assert to_bin(30, [10, 20, 30]) == (20, 30)
    
    assert to_bin(31, [10, 20, 30]) == (30, None)



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(here, 'requirements.txt')) as f:
    install_requires = f.readlines()

setup(
    name='jira-agile-metrics',
    version='0.9',
    description='Agile metrics and summary data extracted from JIRA',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Martin Aspeli',
    author_email='optilude@gmail.com',
    url='https://github.com/optilude/jira-agile-metrics',
    license='MIT',
    keywords='agile jira analytics metrics',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    install_requires=install_requires,
    tests_require=[
        'pytest'
    ],
    entry_points={
        'console_scripts': [
            'jira-agile-metrics=jira_agile_metrics.cli:main',
        ],
    },
)



import logging
import contextlib
import io
import os
import os.path
import shutil
import tempfile
import base64
import zipfile

from flask import Flask, render_template, request
from jira import JIRA, JIRAError

from ..config import config_to_options, CALCULATORS, ConfigError
from ..querymanager import QueryManager
from ..calculator import run_calculators

template_folder = os.path.join(os.path.dirname(__file__), "templates")
static_folder = os.path.join(os.path.dirname(__file__), "static")

app = Flask('jira-agile-metrics',
    template_folder=template_folder,
    static_folder=static_folder
)

logger = logging.getLogger(__name__)

@app.route("/")
def index():
    return render_template('index.html', max_results=request.args.get('max_results', ""))

@app.route("/run", methods=['POST'])
def run():
    config = request.files['config']
    
    data = ""
    has_error = False
    log_buffer = io.StringIO()

    with capture_log(log_buffer, logging.DEBUG, "%(levelname)s: %(message)s"):
        
        # We swallow exceptions here because we want to show them in the output
        # log on the result page.
        try:
            options = config_to_options(config.read())
            override_options(options['connection'], request.form)
        
            # We allow a `max_results` query string parameter for faster debugging
            if request.form.get('max_results'):
                try:
                    options['settings']['max_results'] = int(request.form.get('max_results'))
                except ValueError:
                    options['settings']['max_results'] = None

            jira = get_jira_client(options['connection'])
            query_manager = QueryManager(jira, options['settings'])
            zip_data = get_archive(CALCULATORS, query_manager, options['settings'])
            data = base64.b64encode(zip_data).decode('ascii')
        except Exception as e:
            logger.error("%s", e)
            has_error = True

    return render_template('results.html',
        data=data,
        has_error=has_error,
        log=log_buffer.getvalue()
    )

# Helpers

@contextlib.contextmanager
def capture_log(buffer, level, formatter=None):
    """Temporarily write log output to the StringIO `buffer` with log level
    threshold `level`, before returning logging to normal.
    """
    root_logger = logging.getLogger()
    
    old_level = root_logger.getEffectiveLevel()
    root_logger.setLevel(level)

    handler = logging.StreamHandler(buffer)

    if formatter:
        formatter = logging.Formatter(formatter)
        handler.setFormatter(formatter)

    root_logger.addHandler(handler)

    yield

    root_logger.removeHandler(handler)
    root_logger.setLevel(old_level)

    handler.flush()
    buffer.flush()

def override_options(options, form):
    """Override options from the configuration files with form data where
    applicable.
    """
    for key in options.keys():
        if key in form and form[key] != "":
            options[key] = form[key]

def get_jira_client(connection):
    """Create a JIRA client with the given connection options
    """

    url = connection['domain']
    username = connection['username']
    password = connection['password']
    jira_client_options = connection['jira_client_options']

    jira_options = {'server': url}
    jira_options.update(jira_client_options)

    try:
        return JIRA(jira_options, basic_auth=(username, password))
    except Exception as e:
        if e.status_code == 401:
            raise ConfigError("JIRA authentication failed. Check URL and credentials, and ensure the account is not locked.") from None
        else:
            raise

def get_archive(calculators, query_manager, settings):
    """Run all calculators and write outputs to a temporary directory.
    Create a zip archive of all the files written, and return it as a bytes
    array. Remove the temporary directory on completion.
    """
    zip_data = b''

    cwd = os.getcwd()
    temp_path = tempfile.mkdtemp()

    try:
        os.chdir(temp_path)
        run_calculators(calculators, query_manager, settings)

        with zipfile.ZipFile('metrics.zip', 'w', zipfile.ZIP_STORED) as z:
            for root, dirs, files in os.walk(temp_path):
                for file_name in files:
                    if file_name != 'metrics.zip':
                        z.write(os.path.join(root, file_name), os.path.join('metrics', file_name))
        with open('metrics.zip', 'rb') as metrics_zip:
            zip_data = metrics_zip.read()

    finally:
        os.chdir(cwd)
        shutil.rmtree(temp_path)
    
    return zip_data



import pytest
from pandas import DataFrame, Timestamp

from .cfd import CFDCalculator
from .burnup import BurnupCalculator

from ..utils import extend_dict

@pytest.fixture
def settings(minimal_settings):
    return extend_dict(minimal_settings, {
    })

@pytest.fixture
def query_manager(minimal_query_manager):
    return minimal_query_manager

@pytest.fixture
def results(minimal_cfd_results):
    return extend_dict(minimal_cfd_results, {})

def test_empty(query_manager, settings, cfd_columns):
    results = {
        CFDCalculator: DataFrame([], columns=cfd_columns, index=[])
    }

    calculator = BurnupCalculator(query_manager, settings, results)

    data = calculator.run()
    assert len(data.index) == 0

def test_columns(query_manager, settings, results):
    calculator = BurnupCalculator(query_manager, settings, results)

    data = calculator.run()

    assert list(data.columns) == [
        'Backlog',
        'Done'
    ]

def test_calculate_burnup(query_manager, settings, results):
    calculator = BurnupCalculator(query_manager, settings, results)

    data = calculator.run()

    assert list(data.index) == [
        Timestamp('2018-01-01 00:00:00', freq='D'),
        Timestamp('2018-01-02 00:00:00', freq='D'),
        Timestamp('2018-01-03 00:00:00', freq='D'),
        Timestamp('2018-01-04 00:00:00', freq='D'),
        Timestamp('2018-01-05 00:00:00', freq='D'),
        Timestamp('2018-01-06 00:00:00', freq='D')
    ]

    assert data.to_dict('records') == [
        {'Backlog': 1.0, 'Done': 0.0},
        {'Backlog': 2.0, 'Done': 0.0},
        {'Backlog': 3.0, 'Done': 0.0},
        {'Backlog': 4.0, 'Done': 0.0},
        {'Backlog': 4.0, 'Done': 0.0},
        {'Backlog': 4.0, 'Done': 1.0},
    ]

def test_calculate_burnup_with_different_columns(query_manager, settings, results):
    settings.update({
        'backlog_column': 'Committed',
        'done_column': 'Test'
    })

    calculator = BurnupCalculator(query_manager, settings, results)

    data = calculator.run()

    assert list(data.index) == [
        Timestamp('2018-01-01 00:00:00', freq='D'),
        Timestamp('2018-01-02 00:00:00', freq='D'),
        Timestamp('2018-01-03 00:00:00', freq='D'),
        Timestamp('2018-01-04 00:00:00', freq='D'),
        Timestamp('2018-01-05 00:00:00', freq='D'),
        Timestamp('2018-01-06 00:00:00', freq='D')
    ]

    assert data.to_dict('records') == [
        {'Committed': 0.0, 'Test': 0.0},
        {'Committed': 0.0, 'Test': 0.0},
        {'Committed': 2.0, 'Test': 0.0},
        {'Committed': 3.0, 'Test': 0.0},
        {'Committed': 3.0, 'Test': 1.0},
        {'Committed': 3.0, 'Test': 1.0},
    ]



import pytest
import datetime
import numpy as np
from pandas import DataFrame, DatetimeIndex, Timestamp

from .cycletime import CycleTimeCalculator
from .cfd import CFDCalculator
from .burnup import BurnupCalculator
from .forecast import BurnupForecastCalculator

from ..utils import extend_dict

@pytest.fixture
def settings(minimal_settings):
    return extend_dict(minimal_settings, {
        'burnup_forecast_chart_throughput_window_end': None,
        'burnup_forecast_chart_throughput_window': 8,
        'burnup_forecast_chart_target': 30,
        'burnup_forecast_chart_trials': 10,
        'burnup_forecast_chart_deadline': datetime.date(2018, 1, 30),
        'burnup_forecast_chart_deadline_confidence': 0.85,
        'quantiles': [0.1, 0.3, 0.5],
        'burnup_forecast_chart': 'forecast.png'  # without a file, calculator stops
    })

@pytest.fixture
def query_manager(minimal_query_manager):
    return minimal_query_manager

@pytest.fixture
def results(query_manager, settings, large_cycle_time_results):
    results = large_cycle_time_results.copy()
    results.update({CFDCalculator: CFDCalculator(query_manager, settings, results).run()})
    results.update({BurnupCalculator: BurnupCalculator(query_manager, settings, results).run()})
    return results

def test_empty(query_manager, settings, minimal_cycle_time_columns):
    results = {
        CycleTimeCalculator: DataFrame([], columns=minimal_cycle_time_columns),
        BurnupCalculator: DataFrame([], columns=['Backlog', 'Committed', 'Build', 'Test', 'Done'], index=DatetimeIndex(start=datetime.date(2018, 1, 1), periods=0, freq='D'))
    }

    calculator = BurnupForecastCalculator(query_manager, settings, results)

    data = calculator.run()
    assert data is None

def test_columns(query_manager, settings, results):
    calculator = BurnupForecastCalculator(query_manager, settings, results)

    data = calculator.run()
    assert list(data.columns) == [
        'Trial 0',
        'Trial 1',
        'Trial 2',
        'Trial 3',
        'Trial 4',
        'Trial 5',
        'Trial 6',
        'Trial 7',
        'Trial 8',
        'Trial 9'
    ]

def test_calculate_forecast(query_manager, settings, results):
    calculator = BurnupForecastCalculator(query_manager, settings, results)

    data = calculator.run()

    # because of the random nature of this, we don't know exactly how many records
    # there will be, but will assume at least two
    assert len(data.index) > 0
    assert list(data.index)[0] == Timestamp('2018-01-09 00:00:00', freq='D')
    assert list(data.index)[1] == Timestamp('2018-01-10 00:00:00', freq='D')
    
    for i in range(10):
        trial_values = data['Trial %d' % i]

        # remove na values at the end (not all series will need all dates)
        trial_values = trial_values[:trial_values.last_valid_index()]

        # check that series is monotonically increasing
        trial_diff = np.diff(trial_values)
        assert np.all(trial_diff >= 0)

        # we start with the final value in the burnup, on the final day (2018-01-09)
        assert trial_values[0] == 6

        # we reach the target value
        assert trial_values[-1] == 30

def test_calculate_forecast_settings(query_manager, settings, results):

    settings.update({
        'backlog_column': 'Committed',
        'done_column': 'Test',
        'burnup_forecast_chart_throughput_window_end': datetime.date(2018, 1, 6),
        'burnup_forecast_chart_throughput_window': 4,
        'burnup_forecast_chart_target': None,  # use max of backlog column -- 15
        'burnup_forecast_chart_trials': 10,
        'burnup_forecast_chart_deadline': datetime.date(2018, 1, 30),
        'burnup_forecast_chart_deadline_confidence': 0.85,
        'quantiles': [0.1, 0.3, 0.5]
    })

    results.update({CFDCalculator: CFDCalculator(query_manager, settings, results).run()})
    results.update({BurnupCalculator: BurnupCalculator(query_manager, settings, results).run()})

    calculator = BurnupForecastCalculator(query_manager, settings, results)

    data = calculator.run()

    # because of the random nature of this, we don't know exactly how many records
    # there will be, but will assume at least two
    assert len(data.index) > 0
    assert list(data.index)[0] == Timestamp('2018-01-09 00:00:00', freq='D')
    assert list(data.index)[1] == Timestamp('2018-01-10 00:00:00', freq='D')
    
    for i in range(10):
        trial_values = data['Trial %d' % i]

        # remove na values at the end (not all series will need all dates)
        trial_values = trial_values[:trial_values.last_valid_index()]

        # check that series is monotonically increasing
        trial_diff = np.diff(trial_values)
        assert np.all(trial_diff >= 0)

        # we start with the final value in the burnup, on the final day (2018-01-09)
        assert trial_values[0] == 9

        # we reach the target value
        assert trial_values[-1] == 15



import pytest
from pandas import DataFrame, Timestamp, NaT

from .querymanager import QueryManager
from .utils import extend_dict

from .calculators.cycletime import CycleTimeCalculator
from .calculators.cfd import CFDCalculator

# Fake a portion of the JIRA API

class FauxFieldValue(object):
    """A complex field value, with a name and a typed value
    """
    def __init__(self, name, value):
        self.name = name
        self.value = value

class FauxFields(object):
    """Container for `issue.fields`
    """
    def __init__(self, fields):
        self.__dict__.update(fields)

class FauxChangeItem(object):
    """An item in a changelog change
    """
    def __init__(self, field, fromString, toString):
        self.field = field
        self.from_ = self.fromString = fromString
        self.to = self.toString = toString

class FauxChange(object):
    """A change in a changelog. Contains a list of change items.
    """
    def __init__(self, created, items):
        self.created = created
        self.items = [FauxChangeItem(*i) for i in items]

class FauxChangelog(object):
    """A changelog. Contains a list of changes in `histories`.
    """
    def __init__(self, changes):
        self.histories = changes

class FauxIssue(object):
    """An issue, with a key, change log, and set of fields
    """
    def __init__(self, key, changes, **fields):
        self.key = key
        self.fields = FauxFields(fields)
        self.changelog = FauxChangelog(changes)

class FauxJIRA(object):
    """JIRA interface. Initialised with a set of issues, which will be returned
    by `search_issues()`.
    """

    def __init__(self, fields, issues, options={'server': 'https://example.org'}):
        self._options = options
        self._fields = fields  # [{ id, name }]
        self._issues = issues

    def fields(self):
        return self._fields

    def search_issues(self, jql, *args, **kwargs):
        return self._issues

# Fixtures

@pytest.fixture
def minimal_settings():
    """The smallest `settings` required to build a query manager and cycle time
    calculation.
    """
    return {
        'attributes': {},
        'known_values': {
            'Release': ['R1', 'R3']
        },
        'max_results': None,
        'verbose': False,
        'cycle': [
            {'name': 'Backlog',   'statuses': ['Backlog'],           'type': 'backlog'},
            {'name': 'Committed', 'statuses': ['Next'],              'type': 'accepted'},
            {'name': 'Build',     'statuses': ['Build'],             'type': 'accepted'},
            {'name': 'Test',      'statuses': ['Code review', 'QA'], 'type': 'accepted'},
            {'name': 'Done',      'statuses': ['Done'],              'type': 'complete'}
        ],
        'query_attribute': None,
        'queries': [{'jql': '(filter=123)', 'value': None}],
        
        'backlog_column': 'Backlog',
        'committed_column': 'Committed',
        'final_column': 'Test',
        'done_column': 'Done',
    }


@pytest.fixture
def custom_settings(minimal_settings):
    """A `settings` dict that uses custom fields and attributes.
    """
    return extend_dict(minimal_settings, {
        'attributes': {
            'Release': 'Releases',
            'Team': 'Team',
            'Estimate': 'Size'
        },
        'known_values': {
            'Release': ['R1', 'R3']
        },
    })

# Fields + corresponding columns

@pytest.fixture
def minimal_fields():
    """A `fields` list for all basic fields, but no custom fields.
    """
    return [
        {'id': 'summary',         'name': 'Summary'},
        {'id': 'issuetype',       'name': 'Issue type'},
        {'id': 'status',          'name': 'Status'},
        {'id': 'resolution',      'name': 'Resolution'},
        {'id': 'created',         'name': 'Created date'},
        {'id': 'customfield_100', 'name': 'Flagged'},
    ]

@pytest.fixture
def custom_fields(minimal_fields):
    """A `fields` list with the three custom fields used by `custom_settings`
    """
    return minimal_fields + [
        {'id': 'customfield_001',  'name': 'Team'},
        {'id': 'customfield_002',  'name': 'Size'},
        {'id': 'customfield_003',  'name': 'Releases'},
    ]

@pytest.fixture
def minimal_cycle_time_columns():
    """A columns list for the results of CycleTimeCalculator without any
    custom fields.
    """
    return [
        'key', 'url', 'issue_type', 'summary', 'status', 'resolution',
        'cycle_time', 'completed_timestamp', 'blocked_days', 'impediments',
        'Backlog', 'Committed', 'Build', 'Test', 'Done'
    ]

@pytest.fixture
def custom_cycle_time_columns(minimal_fields):
    """A columns list for the results of CycleTimeCalculator with the three
    custom fields from `custom_settings`.
    """
    return [
        'key', 'url', 'issue_type', 'summary', 'status', 'resolution',
        'Estimate', 'Release', 'Team',
        'cycle_time', 'completed_timestamp', 'blocked_days', 'impediments',
        'Backlog', 'Committed', 'Build', 'Test', 'Done'
    ]

@pytest.fixture
def cfd_columns():
    """A columns list for the results of the CFDCalculator.
    """
    return [
        'Backlog',
        'Committed',
        'Build',
        'Test',
        'Done'
    ]

# Query manager

@pytest.fixture
def minimal_query_manager(minimal_fields, minimal_settings):
    """A minimal query manager (no custom fields)
    """
    jira = FauxJIRA(fields=minimal_fields, issues=[])
    return QueryManager(jira, minimal_settings)

@pytest.fixture
def custom_query_manager(custom_fields, custom_settings):
    """A query manager capable of returning values for custom fields
    """
    jira = FauxJIRA(fields=custom_fields, issues=[])
    return QueryManager(jira, custom_settings)


# Results object with rich cycle time data

def _issues(issues):
    return [{
        'key': 'A-%d' % (idx + 1),
        'url': 'https://example.org/browse/A-%d' % (idx + 1),
        'issue_type': 'Story',
        'summary': 'Generated issue A-%d' % (idx + 1),
        'status': (
            "Done" if i['Done'] is not NaT else
            "Test" if i['Test'] is not NaT else
            "Build" if i['Build'] is not NaT else
            "Committed" if i['Committed'] is not NaT else
            "Backlog"
        ),
        'resoluton': "Done" if i['Done'] is not NaT else None,
        'completed_timestamp': i['Done'] if i['Done'] is not NaT else None,
        'cycle_time': (i['Done'] - i['Committed']) if (i['Done'] is not NaT and i['Committed'] is not NaT) else None,
        'blocked_days': i.get('blocked_days', 0),
        'impediments': i.get('impediments', []),
        'Backlog': i['Backlog'],
        'Committed': i['Committed'],
        'Build': i['Build'],
        'Test': i['Test'],
        'Done': i['Done']
    } for idx, i in enumerate(issues)]

def _ts(datestring, timestring="01:01:01", freq=None):
    return Timestamp('%s %s' % (datestring, timestring,), freq=freq)

@pytest.fixture
def minimal_cycle_time_results(minimal_cycle_time_columns):
    """A results dict mimicing a minimal result from the CycleTimeCalculator.
    """
    return {
        CycleTimeCalculator: DataFrame(_issues([
            dict(Backlog=_ts('2018-01-01'), Committed=NaT,               Build=NaT,               Test=NaT,               Done=NaT),
            dict(Backlog=_ts('2018-01-02'), Committed=_ts('2018-01-03'), Build=NaT,               Test=NaT,               Done=NaT),
            dict(Backlog=_ts('2018-01-03'), Committed=_ts('2018-01-03'), Build=_ts('2018-01-04'), Test=_ts('2018-01-05'), Done=_ts('2018-01-06')),
            dict(Backlog=_ts('2018-01-04'), Committed=_ts('2018-01-04'), Build=NaT,               Test=NaT,               Done=NaT),
        ]), columns=minimal_cycle_time_columns)
    }

@pytest.fixture
def large_cycle_time_results(minimal_cycle_time_columns):
    """A results dict mimicing a larger result from the CycleTimeCalculator.
    """
    return {
        CycleTimeCalculator: DataFrame(_issues([
            # three issues in the backlog
            dict(Backlog=_ts('2018-01-01'), Committed=NaT,               Build=NaT,               Test=NaT,               Done=NaT),
            dict(Backlog=_ts('2018-01-02'), Committed=NaT,               Build=NaT,               Test=NaT,               Done=NaT),
            dict(Backlog=_ts('2018-01-03'), Committed=NaT,               Build=NaT,               Test=NaT,               Done=NaT),

            # three issues started
            dict(Backlog=_ts('2018-01-01'), Committed=_ts('2018-01-02'), Build=NaT,               Test=NaT,               Done=NaT),
            dict(Backlog=_ts('2018-01-01'), Committed=_ts('2018-01-03'), Build=NaT,               Test=NaT,               Done=NaT),
            dict(Backlog=_ts('2018-01-01'), Committed=_ts('2018-01-03'), Build=NaT,               Test=NaT,               Done=NaT),

            # three issues in build
            dict(Backlog=_ts('2018-01-01'), Committed=_ts('2018-01-02'), Build=_ts('2018-01-03'), Test=NaT,               Done=NaT),
            dict(Backlog=_ts('2018-01-01'), Committed=_ts('2018-01-02'), Build=_ts('2018-01-04'), Test=NaT,               Done=NaT),
            dict(Backlog=_ts('2018-01-01'), Committed=_ts('2018-01-02'), Build=_ts('2018-01-04'), Test=NaT,               Done=NaT),

            # three issues in test
            dict(Backlog=_ts('2018-01-01'), Committed=_ts('2018-01-02'), Build=_ts('2018-01-03'), Test=_ts('2018-01-04'), Done=NaT),
            dict(Backlog=_ts('2018-01-01'), Committed=_ts('2018-01-02'), Build=_ts('2018-01-03'), Test=_ts('2018-01-05'), Done=NaT),
            dict(Backlog=_ts('2018-01-01'), Committed=_ts('2018-01-02'), Build=_ts('2018-01-03'), Test=_ts('2018-01-05'), Done=NaT),

            # six issues done, with different cycle times
            dict(Backlog=_ts('2018-01-01'), Committed=_ts('2018-01-02'), Build=_ts('2018-01-03'), Test=_ts('2018-01-04'), Done=_ts('2018-01-07')),
            dict(Backlog=_ts('2018-01-01'), Committed=_ts('2018-01-02'), Build=_ts('2018-01-03'), Test=_ts('2018-01-05'), Done=_ts('2018-01-07')),
            dict(Backlog=_ts('2018-01-01'), Committed=_ts('2018-01-03'), Build=_ts('2018-01-03'), Test=_ts('2018-01-05'), Done=_ts('2018-01-08')),
            dict(Backlog=_ts('2018-01-01'), Committed=_ts('2018-01-03'), Build=_ts('2018-01-03'), Test=_ts('2018-01-04'), Done=_ts('2018-01-08')),
            dict(Backlog=_ts('2018-01-01'), Committed=_ts('2018-01-04'), Build=_ts('2018-01-05'), Test=_ts('2018-01-05'), Done=_ts('2018-01-09')),
            dict(Backlog=_ts('2018-01-01'), Committed=_ts('2018-01-05'), Build=_ts('2018-01-06'), Test=_ts('2018-01-08'), Done=_ts('2018-01-09')),
        ]), columns=minimal_cycle_time_columns)
    }

@pytest.fixture
def minimal_cfd_results(minimal_cycle_time_results, cfd_columns):
    """A results dict mimicing a minimal result from the CycleTimeCalculator.
    """
    return extend_dict(minimal_cycle_time_results, {
        CFDCalculator: DataFrame([
            {'Backlog': 1.0, 'Committed': 0.0, 'Build': 0.0, 'Test': 0.0, 'Done': 0.0},
            {'Backlog': 2.0, 'Committed': 0.0, 'Build': 0.0, 'Test': 0.0, 'Done': 0.0},
            {'Backlog': 3.0, 'Committed': 2.0, 'Build': 0.0, 'Test': 0.0, 'Done': 0.0},
            {'Backlog': 4.0, 'Committed': 3.0, 'Build': 1.0, 'Test': 0.0, 'Done': 0.0},
            {'Backlog': 4.0, 'Committed': 3.0, 'Build': 1.0, 'Test': 1.0, 'Done': 0.0},
            {'Backlog': 4.0, 'Committed': 3.0, 'Build': 1.0, 'Test': 1.0, 'Done': 1.0},
        ], columns=cfd_columns, index=[
            _ts('2018-01-01', '00:00:00', freq='D'),
            _ts('2018-01-02', '00:00:00', freq='D'),
            _ts('2018-01-03', '00:00:00', freq='D'),
            _ts('2018-01-04', '00:00:00', freq='D'),
            _ts('2018-01-05', '00:00:00', freq='D'),
            _ts('2018-01-06', '00:00:00', freq='D')
        ])
    })



import pytest
from pandas import DataFrame

from .cycletime import CycleTimeCalculator
from .throughput import ThroughputCalculator

from ..utils import extend_dict

@pytest.fixture
def settings(minimal_settings):
    return extend_dict(minimal_settings, {
        'throughput_frequency': 'D'
    })

@pytest.fixture
def query_manager(minimal_query_manager):
    return minimal_query_manager

@pytest.fixture
def results(large_cycle_time_results):
    return extend_dict(large_cycle_time_results, {})

def test_empty(query_manager, settings, minimal_cycle_time_columns):
    results = {
        CycleTimeCalculator: DataFrame([], columns=minimal_cycle_time_columns, index=[])
    }

    calculator = ThroughputCalculator(query_manager, settings, results)

    data = calculator.run()
    assert list(data.columns) == ['count']
    assert len(data.index) == 0


def test_columns(query_manager, settings, results):
    calculator = ThroughputCalculator(query_manager, settings, results)

    data = calculator.run()

    assert list(data.columns) == ['count']

def test_calculate_throughput(query_manager, settings, results):
    calculator = ThroughputCalculator(query_manager, settings, results)

    data = calculator.run()

    assert data.to_dict('records') == [{'count': 2}, {'count': 2}, {'count': 2}]
